
using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Threading;
using ClarionSimulation;
using WorldServerLibrary;
using WorldServerLibrary.Model;
using WorldServerLibrary.Exceptions;

namespace ClarionDEMO
{
	class MainClass
	{

		private double D90  = 1.5705;
		private double D180 = 3.1415;
		private double D270 = 4.7120;
		private double D360 = 6.2830;
		private double D10  = 0.1745;
	
		private double JEWELS_TO_GENERATE = 6;
		
		private double FOOD_TO_GENERATE = 4;
		
		private int PROX_LIMIT = 30;
		
		
		#region properties
		private WorldServer worldServer = null;
        private ClarionAgent agent;
        String creatureId = String.Empty;
        String creatureName = String.Empty;
		double prad = 0;
		
		Creature whichCreature = null;
		
		Thing deliverySpot = null;
		
		#endregion

		#region constructor
		public MainClass() {
			Console.WriteLine ("Clarion WorldServer3D Client V0.1 - Eduardo Seiti de Oliveira - RA 940011");
			
			agent = new ClarionAgent();
            agent.OnNewVisualSensorialInformation += new InputVisualSensorialInformationEventHandler(agent_OnNewVisualSensorialInformation);
            agent.OnNewExternalActionSelected += new OutputActionChunkEventHandler(agent_OnNewExternalActionSelected);
			try
            {
//                worldServer = new WorldServer("192.168.0.164", 4011);

				worldServer = new WorldServer("localhost", 4011);
                
				String message = worldServer.Connect();

                if (worldServer != null && worldServer.IsConnected)
                {
                    Console.Out.WriteLine ("[SUCCESS] " + message + "\n");
                    worldServer.NewCreature (400, 200, 0, out creatureId, out creatureName);
                    worldServer.NewBrick (4, 800,   0, 802, 600);
                    worldServer.NewBrick (4,  -2,   0,   0, 600);
                    worldServer.NewBrick (4,  -2,  -2, 802,   0);
                    worldServer.NewBrick (4,  -2, 600, 800, 602);

                    if (!String.IsNullOrWhiteSpace(creatureId))
                    {
                        worldServer.SendStartCamera (creatureId);
                        worldServer.SendStartCreature (creatureId);
                    }

                    Console.Out.WriteLine ("Creature created with name: " + creatureId + "\n");
					
					
					// Create the leaflets list
					
					String leafletsList = worldServer.SendCreateLeaflets();
					
					
					// Creates 1 jewel of each color
					
					worldServer.NewJewel (0, 100, 100);
					worldServer.NewJewel (1, 700, 500);
					worldServer.NewJewel (2, 100, 500);
					worldServer.NewJewel (3, 700, 100);
					worldServer.NewJewel (4, 100, 200);
					worldServer.NewJewel (5, 600, 200);
					
					
					// Create some food
					
					worldServer.NewFood (1, 300, 300);
					worldServer.NewFood (1, 500, 300);
					worldServer.NewFood (1, 100, 300);
					worldServer.NewFood (1, 400, 500);
					
					
					// Get the current delivery spot
					
					deliverySpot = worldServer.SendGetSimulPars();
					
					Console.WriteLine ("Created leaflets: {0}", leafletsList);
					
                    agent.Run ();
					Console.Out.WriteLine ("Running Simulation ...\n");
                }
            }
            catch (WorldServerInvalidArgument invalidArtgument)
            {
                Console.Out.WriteLine(String.Format("[ERROR] Invalid Argument: {0}\n", invalidArtgument.Message));
            }
            catch (WorldServerConnectionError serverError)
            {
                Console.Out.WriteLine(String.Format("[ERROR] Is is not possible to connect to server: {0}\n", serverError.Message));
            }
            catch (Exception ex)
            {
                Console.Out.WriteLine(String.Format("[ERROR] Unknown Error: {0}\n", ex.Message));
            }
		}
		#endregion

		#region Methods
		public static void Main (string[] args)	{
			new MainClass();
		}

        IList<Thing> agent_OnNewVisualSensorialInformation()
        {
            IList<Thing> response = null;
			
			Thread.CurrentThread.CurrentCulture = new CultureInfo("en-US");
			
            if (worldServer != null && worldServer.IsConnected)
            {
                response = worldServer.SendGetCreatureState(creatureName);
				
				whichCreature = (Creature) response.First();
				
				prad = (Math.PI / 180) * whichCreature.Pitch;
				
			    while (prad > Math.PI) prad -= 2 * Math.PI;
			    while (prad < - Math.PI) prad += 2 * Math.PI;
				
				// Always send the delivery spot...
				
				response.Add (deliverySpot);

            }

            return response;
        }
		
		
		
		void createThingsInWorld ()
		{
			Random generator = new Random();
			
			List<int> genX = new List<int>();
			List<int> genY = new List<int>();
			
			int newX;
			int newY;
			
			IEnumerable<int> nearX;
			IEnumerable<int> nearY;
			
			for (int i = 0; i < JEWELS_TO_GENERATE; i++)
			{
				do
				{
					newX = generator.Next (50, 750);
					newY = generator.Next (50, 550);
					
					nearX = genX.Where (item => item <= newX + PROX_LIMIT && item >= newX - PROX_LIMIT);
					nearY = genY.Where (item => item <= newY + PROX_LIMIT && item >= newY - PROX_LIMIT);
				}
				while (nearX.Count () > 0 || nearY.Count () > 0);

				genX.Add (newX);
				genY.Add (newY);
				
				worldServer.NewJewel (i % 6, newX, newY);
			}
			
			for (int i = 0; i < FOOD_TO_GENERATE; i++)
			{
				do
				{
					newX = generator.Next (50, 750);
					newY = generator.Next (50, 550);
					
					nearX = genX.Where (item => item <= newX + PROX_LIMIT && item >= newX - PROX_LIMIT);
					nearY = genY.Where (item => item <= newY + PROX_LIMIT && item >= newY - PROX_LIMIT);
				}
				while (nearX.Count () > 0 || nearY.Count () > 0);

				genX.Add (newX);
				genY.Add (newY);
				
				worldServer.NewFood (1, newX, newY);
			}			
		}
		
		
		
        void agent_OnNewExternalActionSelected(ClarionAgentActionType externalAction, Thing whichThing)
        {   
			
			Console.WriteLine ("No delegate: - externalAction: {0} - whichThing: {1}", externalAction, whichThing != null ? whichThing.ToString() : null);
			
			
			Thread.CurrentThread.CurrentCulture = new CultureInfo("en-US");
			
            if (worldServer != null && worldServer.IsConnected)
            {
                switch (externalAction)
                {
                    case ClarionAgentActionType.DO_NOTHING:
                        // Do nothing as the own value says
                        break;
					
                    case ClarionAgentActionType.ROTATE_CLOCKWISE:
					    double newprad = prad + Math.PI/100;
					
					    while (newprad > Math.PI) newprad -= 2 * Math.PI;
			            while (newprad < - Math.PI) newprad += 2 * Math.PI;
					
                        worldServer.SendSetAngle (creatureId, 0, 0, newprad);
					
                        break;
										
					case ClarionAgentActionType.COLLECT_JEWEL:
					
						worldServer.SendSackIt (creatureId, whichThing.Name);
					
						// Stop the creature
					
						worldServer.SendSetTurn (creatureId, 0, 0, 0);
					
						break;
					
					
					case ClarionAgentActionType.EAT_FOOD:
					
						worldServer.SendEatIt (creatureId, whichThing.Name);

						// Stop the creature
					
						worldServer.SendSetTurn (creatureId, 0, 0, 0);
					
						break;
					
					
					case ClarionAgentActionType.CONTOUR_OBSTACLE:
						
						double reboundAngle = 1.5705;
						double wheelSpeed = 1;
					
						if (whichThing != null)
						{					
							double distX = Math.Min (Math.Abs (whichThing.X1 - whichCreature.X1), Math.Abs (whichThing.X2 - whichCreature.X2));
							double distY = Math.Min (Math.Abs (whichThing.Y1 - whichCreature.Y1), Math.Abs (whichThing.Y2 - whichCreature.Y2));
						
							if (((prad >=  D90 - D10) && (prad <=  D90 + D10)) ||
							    ((prad >= D180 - D10) && (prad <= D180 + D10)) ||
							    ((prad >= D270 - D10) && (prad <= D270 + D10)) ||
							    ((prad >= D360 - D10) && (prad <= 0.0 + D10)))
							{
								reboundAngle = 0.0;
								wheelSpeed = -1;
							}
							else
							{										
								if ((prad > 1.5705 && prad < 3.1415) || prad > 4.7120)
								{
									reboundAngle = -reboundAngle;
								}
							
							
								if (distX > distY)
								{
									// Go counter-clockwise....
								
									reboundAngle = -reboundAngle;
								}
							}
						}
						else
						{
							// There are multiple things to contour; just go back a while
							
							reboundAngle = 0.0;
							wheelSpeed = -1;
						}
					
						worldServer.SendSetAngle (creatureId, wheelSpeed, wheelSpeed, prad + reboundAngle);
						
						break;
					
					
					case ClarionAgentActionType.GO_POSITION:
					
						worldServer.SendSetGoTo (creatureId, 1, 1, whichThing.X1, whichThing.Y1);
					
						break;
					
					case ClarionAgentActionType.DELIVER_LEAFLET:
					
						worldServer.SendDeliverLeaflet (creatureId, ((Leaflet) whichThing).ID);
					
						break;
					
					
					case ClarionAgentActionType.CREATE_THINGS:
						
						createThingsInWorld();
						
						break;
					
					
                    default:
                        break;
                }
            }
        }

        #endregion
	}
	
	
}
